# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import math
from itertools import chain
from pathlib import Path
from typing import Optional
from typing import Union

import torch

from anemoi.utils.humanize import bytes
from anemoi.utils.text import table


class GraphDescriptor:
    """Class for descripting the graph."""

    def __init__(self, path: Union[str, Path], **kwargs):
        self.path = path
        self.graph = torch.load(self.path, weights_only=False, map_location="cpu")

    @property
    def total_size(self):
        """Total size of the tensors in the graph (in bytes)."""
        total_size = 0

        for store in chain(self.graph.node_stores, self.graph.edge_stores):
            for value in store.values():
                if isinstance(value, torch.Tensor):
                    total_size += value.numel() * value.element_size()

        return total_size

    def get_node_summary(self) -> list[list]:
        """Summary of the nodes in the graph.

        Returns
        -------
        list[list]
            Returns a list for each subgraph with the following information:
            - Node name.
            - Number of nodes.
            - List of attribute names.
            - Total dimension of the attributes.
            - Min. latitude.
            - Max. latitude.
            - Min. longitude.
            - Max. longitude.
        """
        node_summary = []
        for name, nodes in self.graph.node_items():
            attributes = nodes.node_attrs()
            attributes.remove("x")

            node_summary.append(
                [
                    name,
                    nodes.num_nodes,
                    ", ".join(attributes),
                    sum(nodes[attr].shape[1] for attr in attributes if isinstance(nodes[attr], torch.Tensor)),
                    nodes.x[:, 0].min().item() / 2 / math.pi * 360,
                    nodes.x[:, 0].max().item() / 2 / math.pi * 360,
                    nodes.x[:, 1].min().item() / 2 / math.pi * 360,
                    nodes.x[:, 1].max().item() / 2 / math.pi * 360,
                ]
            )
        return node_summary

    def get_edge_summary(self) -> list[list]:
        """Summary of the edges in the graph.

        Returns
        -------
        list[list]
            Returns a list for each subgraph with the following information:
            - Source node name.
            - Destination node name.
            - Number of edges.
            - Number of isolated source nodes.
            - Number of isolated target nodes.
            - Total dimension of the attributes.
            - List of attribute names.
        """
        edge_summary = []
        for (src_name, _, dst_name), edges in self.graph.edge_items():
            attributes = edges.edge_attrs()
            attributes.remove("edge_index")

            edge_summary.append(
                [
                    src_name,
                    dst_name,
                    edges.num_edges,
                    self.graph[src_name].num_nodes - len(torch.unique(edges.edge_index[0])),
                    self.graph[dst_name].num_nodes - len(torch.unique(edges.edge_index[1])),
                    sum(edges[attr].shape[1] for attr in attributes),
                    ", ".join([f"{attr}({edges[attr].shape[1]}D)" for attr in attributes]),
                ]
            )
        return edge_summary

    def get_node_attribute_table(self) -> list[list]:
        node_attributes = []
        for node_name, node_store in self.graph.node_items():
            node_attr_names = node_store.node_attrs()
            node_attr_names.remove("x")  # Remove the coordinates from statistics table
            for node_attr_name in node_attr_names:
                node_attributes.append(
                    [
                        "Node",
                        node_name,
                        node_attr_name,
                        node_store[node_attr_name].dtype,
                        node_store[node_attr_name].float().min().item(),
                        node_store[node_attr_name].float().mean().item(),
                        node_store[node_attr_name].float().max().item(),
                        node_store[node_attr_name].float().std().item(),
                    ]
                )
        return node_attributes

    def get_edge_attribute_table(self) -> list[list]:
        edge_attributes = []
        for (source_name, _, target_name), edge_store in self.graph.edge_items():
            edge_attr_names = edge_store.edge_attrs()
            edge_attr_names.remove("edge_index")  # Remove the edge index from statistics table
            for edge_attr_name in edge_attr_names:
                edge_attributes.append(
                    [
                        "Edge",
                        f"{source_name}-->{target_name}",
                        edge_attr_name,
                        edge_store[edge_attr_name].dtype,
                        edge_store[edge_attr_name].float().min().item(),
                        edge_store[edge_attr_name].float().mean().item(),
                        edge_store[edge_attr_name].float().max().item(),
                        edge_store[edge_attr_name].float().std().item(),
                    ]
                )

        return edge_attributes

    def get_attribute_table(self) -> list[list]:
        """Get a table with the attributes of the graph."""
        attribute_table = []
        attribute_table.extend(self.get_node_attribute_table())
        attribute_table.extend(self.get_edge_attribute_table())
        return attribute_table

    def describe(self, show_attribute_distributions: Optional[bool] = True) -> None:
        """Describe the graph."""
        print()
        print(f"📦 Path       : {self.path}")
        print(f"💽 Size       : {bytes(self.total_size)} ({self.total_size})")
        print()
        print("🪩  Nodes summary")
        print()
        print(
            table(
                self.get_node_summary(),
                header=[
                    "Nodes name",
                    "Num. nodes",
                    "Attributes",
                    "Attribute dim",
                    "Min. latitude",
                    "Max. latitude",
                    "Min. longitude",
                    "Max. longitude",
                ],
                align=["<", ">", ">", ">", ">", ">", ">", ">"],
                margin=3,
            )
        )
        print()
        print()
        print("🌐  Edges summary")
        print()
        print(
            table(
                self.get_edge_summary(),
                header=[
                    "Source",
                    "Target",
                    "Num. edges",
                    "Isolated Source",
                    "Isolated Target",
                    "Attribute dim",
                    "Attributes",
                ],
                align=["<", "<", ">", ">", ">", ">", ">"],
                margin=3,
            )
        )
        print()
        if show_attribute_distributions:
            print()
            print("📊 Attribute distributions")
            print()
            print(
                table(
                    self.get_attribute_table(),
                    header=[
                        "Type",
                        "Source",
                        "Name",
                        "Dtype",
                        "Min.",
                        "Mean",
                        "Max.",
                        "Std. dev.",
                    ],
                    align=["<", "<", ">", ">", ">", ">", ">", ">"],
                    margin=3,
                )
            )
            print()
        print("🔋 Graph ready.")
        print()



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import numpy as np
import pytest
import torch
import yaml
from torch_geometric.data import HeteroData

lats = [-0.15, 0, 0.15]
lons = [0, 0.25, 0.5, 0.75]


class MockZarrDataset:
    """Mock Zarr dataset with latitudes and longitudes attributes."""

    def __init__(self, latitudes, longitudes, grids=None):
        self.latitudes = latitudes
        self.longitudes = longitudes
        self.num_nodes = len(latitudes)
        if grids is not None:
            self.grids = grids


@pytest.fixture
def mock_zarr_dataset() -> MockZarrDataset:
    """Mock zarr dataset with nodes."""
    coords = 2 * torch.pi * np.array([[lat, lon] for lat in lats for lon in lons])
    return MockZarrDataset(latitudes=coords[:, 0], longitudes=coords[:, 1])


@pytest.fixture
def mock_zarr_dataset_cutout() -> MockZarrDataset:
    """Mock zarr dataset with nodes."""
    coords = 2 * torch.pi * np.array([[lat, lon] for lat in lats for lon in lons])
    grids = int(0.3 * len(coords)), len(coords) - int(0.3 * len(coords))
    return MockZarrDataset(latitudes=coords[:, 0], longitudes=coords[:, 1], grids=grids)


@pytest.fixture
def mock_grids_path(tmp_path) -> tuple[str, int]:
    """Mock grid_definition_path with files for 3 resolutions."""
    num_nodes = len(lats) * len(lons)
    for resolution in ["o16", "o48", "5km5"]:
        file_path = tmp_path / f"grid-{resolution}.npz"
        np.savez(file_path, latitudes=np.random.rand(num_nodes), longitudes=np.random.rand(num_nodes))
    return str(tmp_path), num_nodes


@pytest.fixture
def graph_with_nodes() -> HeteroData:
    """Graph with 12 nodes over the globe, stored in \"test_nodes\"."""
    coords = np.array([[lat, lon] for lat in lats for lon in lons])
    graph = HeteroData()
    graph["test_nodes"].x = 2 * torch.pi * torch.tensor(coords)
    graph["test_nodes"].mask = torch.tensor([True] * len(coords)).unsqueeze(-1)
    graph["test_nodes"].mask2 = torch.tensor([True] * (len(coords) - 2) + [False] * 2).unsqueeze(-1)
    graph["test_nodes"]["_grid_reference_distance"] = 0.75
    return graph


@pytest.fixture
def graph_with_isolated_nodes() -> HeteroData:
    graph = HeteroData()
    graph["test_nodes"].x = torch.tensor([[1], [2], [3], [4], [5], [6]])
    graph["test_nodes"]["mask_attr"] = torch.tensor([[1], [1], [1], [0], [0], [0]], dtype=torch.bool)
    graph["test_nodes", "to", "test_nodes"].edge_index = torch.tensor([[2, 3, 4], [1, 2, 3]])
    return graph


@pytest.fixture
def graph_nodes_and_edges() -> HeteroData:
    """Graph with 1 set of nodes and edges."""
    coords = np.array([[lat, lon] for lat in lats for lon in lons])
    graph = HeteroData()
    graph["test_nodes"].x = 2 * torch.pi * torch.tensor(coords)
    graph["test_nodes"].mask = torch.tensor([True] * len(coords)).unsqueeze(-1)
    graph[("test_nodes", "to", "test_nodes")].edge_index = torch.tensor([[3, 1, 2, 0], [2, 0, 1, 3]])
    graph[("test_nodes", "to", "test_nodes")].edge_attr = (
        10 * graph[("test_nodes", "to", "test_nodes")].edge_index[0][:, None]
    )
    return graph


@pytest.fixture
def graph_long_and_short_edges() -> HeteroData:
    """Graph with a pair of short (800km) and a pair of long (20000km) edges."""
    graph = HeteroData()
    graph["test_nodes"].x = 2 * torch.pi * torch.tensor([[-0.01, 0], [0.01, 0], [-0.01, 0.5], [0.01, 0.5]])
    graph["test_nodes"]["southern_hemisphere_mask"] = torch.tensor([[1], [0], [1], [0]], dtype=torch.bool)
    graph["test_nodes", "to", "test_nodes"].edge_index = torch.tensor([[0, 0, 1, 3], [1, 3, 2, 2]])
    return graph


@pytest.fixture
def config_file(tmp_path) -> tuple[str, str]:
    """Mock grid_definition_path with files for 3 resolutions."""
    cfg = {
        "nodes": {
            "test_nodes": {
                "node_builder": {
                    "_target_": "anemoi.graphs.nodes.NPZFileNodes",
                    "npz_file": str(tmp_path) + "/grid-o16.npz",
                },
            },
        },
        "edges": [
            {
                "source_name": "test_nodes",
                "target_name": "test_nodes",
                "edge_builders": [
                    {"_target_": "anemoi.graphs.edges.KNNEdges", "num_nearest_neighbours": 3},
                ],
                "attributes": {
                    "dist_norm": {"_target_": "anemoi.graphs.edges.attributes.EdgeLength"},
                    "edge_dirs": {"_target_": "anemoi.graphs.edges.attributes.EdgeDirection"},
                },
            },
        ],
    }
    file_name = "config.yaml"

    with (tmp_path / file_name).open("w") as file:
        yaml.dump(cfg, file)

    return tmp_path, file_name



# (C) Copyright 2024- Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import pytest
import torch
from torch_geometric.data import HeteroData

from anemoi.graphs.nodes.attributes import CutOutMask


def test_cutout_mask(mocker, graph_with_nodes: HeteroData, mock_zarr_dataset_cutout):
    """Test attribute builder for CutOutMask."""
    # Add dataset attribute required by CutOutMask
    graph_with_nodes["test_nodes"]["_dataset"] = {"cutout": None}

    mocker.patch("anemoi.graphs.nodes.attributes.masks.open_dataset", return_value=mock_zarr_dataset_cutout)
    mask = CutOutMask().compute(graph_with_nodes, "test_nodes")

    assert mask is not None
    assert isinstance(mask, torch.Tensor)
    assert mask.dtype == torch.bool
    assert mask.shape[0] == graph_with_nodes["test_nodes"].x.shape[0]


def test_cutout_mask_missing_dataset(graph_with_nodes: HeteroData):
    """Test CutOutMask fails when dataset attribute is missing."""
    node_attr_builder = CutOutMask()
    with pytest.raises(AssertionError):
        node_attr_builder.compute(graph_with_nodes, "test_nodes")


def test_cutout_mask_missing_cutout(graph_with_nodes: HeteroData):
    """Test CutOutMask fails when cutout key is missing."""
    graph_with_nodes["test_nodes"]["_dataset"] = {}

    node_attr_builder = CutOutMask()
    with pytest.raises(AssertionError):
        node_attr_builder.compute(graph_with_nodes, "test_nodes")



# (C) Copyright 2024- Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from __future__ import annotations

import logging

import numpy as np
import torch
from scipy.spatial import ConvexHull
from scipy.spatial import SphericalVoronoi
from scipy.spatial import Voronoi
from torch_geometric.data.storage import NodeStorage

from anemoi.graphs.generate.transforms import latlon_rad_to_cartesian
from anemoi.graphs.nodes.attributes.base_attributes import BaseNodeAttribute

LOGGER = logging.getLogger(__name__)


class UniformWeights(BaseNodeAttribute):
    """Implements a uniform weight for the nodes.

    Methods
    -------
    compute(self, graph, nodes_name)
        Compute the area attributes for each node.
    """

    def get_raw_values(self, nodes: NodeStorage, **kwargs) -> torch.Tensor:
        """Compute the weights.

        Parameters
        ----------
        nodes : NodeStorage
            Nodes of the graph.
        kwargs : dict
            Additional keyword arguments.

        Returns
        -------
        torch.Tensor
            Ones.
        """
        return torch.ones((nodes.num_nodes,))


class AreaWeights(BaseNodeAttribute):
    """Implements the area of the nodes as the weights.

    Attributes
    ----------
    flat: bool
        If True, the area is computed in 2D, otherwise in 3D.
    **other: Any
        Additional keyword arguments, see PlanarAreaWeights and SphericalAreaWeights
        for details.

    Methods
    -------
    compute(self, graph, nodes_name)
        Compute the area attributes for each node.
    """

    def __new__(cls, flat: bool = False, **kwargs):
        logging.warning(
            "Creating %s with flat=%s and kwargs=%s. In a future release, AreaWeights will be deprecated: please use directly PlanarAreaWeights or SphericalAreaWeights.",
            cls.__name__,
            flat,
            kwargs,
        )
        if flat:
            return PlanarAreaWeights(**kwargs)
        return SphericalAreaWeights(**kwargs)


class PlanarAreaWeights(BaseNodeAttribute):
    """Implements the 2D area of the nodes as the weights.

    Attributes
    ----------
    norm : str
        Normalisation of the weights.

    Methods
    -------
    compute(self, graph, nodes_name)
        Compute the area attributes for each node.
    """

    def _compute_mean_nearest_distance(self, points: np.ndarray) -> float:
        """Compute mean distance to nearest neighbor for each point.

        Parameters
        ----------
        points : np.ndarray
            Array of point coordinates (N x 2)

        Returns
        -------
        float
            Mean nearest neighbor distance
        """
        from scipy.spatial import cKDTree

        tree = cKDTree(points)
        distances, _ = tree.query(points, k=2)
        return float(distances[:, 1].mean())

    def _get_boundary_ring(self, points: np.ndarray, resolution: float) -> np.ndarray:
        """Add a ring of boundary points around the input points.

        Parameters
        ----------
        points : np.ndarray
            Original point coordinates
        resolution : float
            Approximate spacing between points

        Returns
        -------
        np.ndarray
            Array including original and boundary points
        """
        # Get convex hull vertices
        hull = ConvexHull(points)
        hull_points = points[hull.vertices]

        # Expand hull outward
        centroid = np.mean(hull_points, axis=0)
        vectors = hull_points - centroid
        expanded_hull = hull_points + vectors * (2**0.5 * resolution / np.linalg.norm(vectors, axis=1)[:, np.newaxis])

        # Create points along each hull edge
        boundary_points = []
        p1 = expanded_hull
        p2 = np.roll(expanded_hull, 1, axis=0)

        # Calculate number of points needed along this edge
        edge_length = np.linalg.norm(p2 - p1, axis=1)
        num_points = np.ceil(edge_length / resolution).astype(int)

        for i in np.where(num_points > 2)[0]:
            # Create evenly spaced points along the edge
            t = np.linspace(0, 1, num_points[i])[1:-1][:, None]  # Exclude last point to avoid duplicates
            edge_points = p1[i] + t * (p2[i] - p1[i])
            boundary_points.append(edge_points)

        return np.concatenate([expanded_hull, np.vstack(boundary_points)])

    def get_raw_values(self, nodes: NodeStorage, **kwargs) -> torch.Tensor:
        points = nodes.x.cpu().numpy()
        resolution = self._compute_mean_nearest_distance(points)
        boundary_points = self._get_boundary_ring(points, resolution)

        # Compute convex hull over all points (boundary ring included)
        extended_points = np.vstack([points, boundary_points])
        v = Voronoi(extended_points, qhull_options="QJ Pp")

        # Compute the area of each node's region, excluding those in the boundary ring
        areas = []
        for idx in range(len(points)):
            p_idx = v.point_region[idx]
            r = v.regions[p_idx]
            poly_coords = v.vertices[r]
            area = ConvexHull(poly_coords).volume
            areas.append(area)

        return torch.from_numpy(np.array(areas))


class SphericalAreaWeights(BaseNodeAttribute):
    """Implements the 3D area of the nodes as the weights.

    Attributes
    ----------
    norm : str
        Normalisation of the weights.
    radius : float
        Radius of the sphere.
    centre : np.ndarray
        Centre of the sphere.
    fill_value : float
        Value to fill the empty regions.

    Methods
    -------
    compute(self, graph, nodes_name)
        Compute the area attributes for each node.
    """

    def __init__(
        self,
        norm: str | None = None,
        radius: float = 1.0,
        centre: np.ndarray = np.array([0, 0, 0]),
        fill_value: float = 0.0,
        dtype: str = "float32",
    ) -> None:
        assert isinstance(fill_value, float) or isinstance(
            fill_value, int
        ), f"fill_value must be float or nan but it is {type(fill_value)}"
        assert (
            isinstance(radius, float) or isinstance(radius, int)
        ) and radius > 0, f"radius must be a positive value, but radius={radius}"
        super().__init__(norm, dtype)
        self.radius = radius
        self.centre = centre
        self.fill_value = fill_value

    def get_raw_values(self, nodes: NodeStorage, **kwargs) -> np.ndarray:
        """Compute the area associated to each node.

        It uses Voronoi diagrams to compute the area of each node.

        Parameters
        ----------
        nodes : NodeStorage
            Nodes of the graph.
        kwargs : dict
            Additional keyword arguments.

        Returns
        -------
        np.ndarray
            Attributes.
        """
        points = latlon_rad_to_cartesian(nodes.x.cpu())
        sv = SphericalVoronoi(points, self.radius, self.centre)
        mask = np.array([bool(i) for i in sv.regions])
        sv.regions = [region for region in sv.regions if region]
        # compute the area weight without empty regions
        area_weights = sv.calculate_areas()
        if (null_nodes := (~mask).sum()) > 0:
            LOGGER.warning(
                "%s is filling %d (%.2f%%) nodes with value %f",
                self.__class__.__name__,
                null_nodes,
                100 * null_nodes / len(mask),
                self.fill_value,
            )
        result = np.ones(points.shape[0]) * self.fill_value
        result[mask] = area_weights
        LOGGER.debug(
            "There are %d of weights, which (unscaled) add up a total weight of %.2f.",
            len(result),
            result.sum(),
        )
        return torch.from_numpy(result)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from __future__ import annotations

import logging
import warnings
from importlib.util import find_spec

import numpy as np
import torch
from sklearn.neighbors import NearestNeighbors
from torch_geometric.data.storage import NodeStorage

from anemoi.graphs.edges.builders.base import BaseEdgeBuilder
from anemoi.graphs.edges.builders.masking import NodeMaskingMixin

LOGGER = logging.getLogger(__name__)

TORCH_CLUSTER_AVAILABLE = find_spec("torch_cluster") is not None


class KNNEdges(BaseEdgeBuilder, NodeMaskingMixin):
    """Computes KNN based edges and adds them to the graph.

    Attributes
    ----------
    source_name : str
        The name of the source nodes.
    target_name : str
        The name of the target nodes.
    num_nearest_neighbours : int
        Number of nearest neighbours.
    source_mask_attr_name : str | None
        The name of the source mask attribute to filter edge connections.
    target_mask_attr_name : str | None
        The name of the target mask attribute to filter edge connections.

    Methods
    -------
    register_edges(graph)
        Register the edges in the graph.
    register_attributes(graph, config)
        Register attributes in the edges of the graph.
    update_graph(graph, attrs_config)
        Update the graph with the edges.
    """

    def __init__(
        self,
        source_name: str,
        target_name: str,
        num_nearest_neighbours: int,
        source_mask_attr_name: str | None = None,
        target_mask_attr_name: str | None = None,
    ) -> None:
        super().__init__(source_name, target_name, source_mask_attr_name, target_mask_attr_name)
        assert isinstance(num_nearest_neighbours, int), "Number of nearest neighbours must be an integer."
        assert num_nearest_neighbours > 0, "Number of nearest neighbours must be positive."
        self.num_nearest_neighbours = num_nearest_neighbours

    def _compute_edge_index_pyg(self, source_nodes: NodeStorage, target_nodes: NodeStorage) -> torch.Tensor:
        from torch_cluster.knn import knn

        source_coords, target_coords = self.get_cartesian_node_coordinates(source_nodes, target_nodes)

        edge_index = knn(source_coords, target_coords, k=self.num_nearest_neighbours)

        return torch.flip(edge_index, [0])

    def _compute_edge_index_sklearn(self, source_nodes: NodeStorage, target_nodes: NodeStorage) -> torch.Tensor:
        source_coords, target_coords = self.get_cartesian_node_coordinates(source_nodes, target_nodes)
        nearest_neighbour = NearestNeighbors(metric="euclidean", n_jobs=4)
        nearest_neighbour.fit(source_coords.cpu())
        adj_matrix = nearest_neighbour.kneighbors_graph(
            target_coords.cpu(),
            n_neighbors=self.num_nearest_neighbours,
        ).tocoo()

        # Post-process the adjacency matrix. Add masked nodes.
        adj_matrix = self.undo_masking(adj_matrix, source_nodes, target_nodes)
        edge_index = torch.from_numpy(np.stack([adj_matrix.col, adj_matrix.row], axis=0))
        return edge_index

    def compute_edge_index(self, source_nodes: NodeStorage, target_nodes: NodeStorage) -> torch.Tensor:
        """Compute the edge indices for the KNN method.

        Parameters
        ----------
        source_nodes : NodeStorage
            The source nodes.
        target_nodes : NodeStorage
            The target nodes.

        Returns
        -------
        torch.Tensor of shape (2, num_edges)
            Indices of source and target nodes connected by an edge.
        """
        assert self.num_nearest_neighbours is not None, "number of neighbors required for knn encoder"
        LOGGER.info(
            "Using KNN-Edges (with %d nearest neighbours) between %s and %s.",
            self.num_nearest_neighbours,
            self.source_name,
            self.target_name,
        )

        if TORCH_CLUSTER_AVAILABLE:
            edge_index = self._compute_edge_index_pyg(source_nodes, target_nodes)
        else:
            warnings.warn(
                "The 'torch-cluster' library is not installed. Installing 'torch-cluster' can significantly improve "
                "performance for graph creation. You can install it using 'pip install torch-cluster'.",
                UserWarning,
            )
            edge_index = self._compute_edge_index_sklearn(source_nodes, target_nodes)

        return edge_index



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from __future__ import annotations

from collections.abc import Iterable

import networkx as nx
import numpy as np
import trimesh
from sklearn.neighbors import BallTree

from anemoi.graphs.generate.masks import KNNAreaMaskBuilder
from anemoi.graphs.generate.transforms import cartesian_to_latlon_rad
from anemoi.graphs.generate.utils import get_coordinates_ordering


def create_tri_nodes(
    resolution: int, area_mask_builder: KNNAreaMaskBuilder | None = None
) -> tuple[nx.DiGraph, np.ndarray, list[int]]:
    """Creates a global mesh from a refined icosahedron.

    This method relies on the trimesh python library.

    Parameters
    ----------
    resolution : int
        Level of mesh resolution to consider.
    area_mask_builder : KNNAreaMaskBuilder
        KNNAreaMaskBuilder with the cloud of points to limit the mesh area, by default None.

    Returns
    -------
    graph : networkx.Graph
        The specified graph (only nodes) sorted by latitude and longitude.
    coords_rad : np.ndarray
        The node coordinates (not ordered) in radians.
    node_ordering : list[int]
        Order of the node coordinates to be sorted by latitude and longitude.
    """
    coords_rad = get_latlon_coords_icosphere(resolution)

    node_ordering = get_coordinates_ordering(coords_rad)

    if area_mask_builder is not None:
        area_mask = area_mask_builder.get_mask(coords_rad)
        node_ordering = node_ordering[area_mask[node_ordering]]

    # Creates the graph, with the nodes sorted by latitude and longitude.
    nx_graph = create_nx_graph_from_tri_coords(coords_rad, node_ordering)

    return nx_graph, coords_rad, list(node_ordering)


def create_stretched_tri_nodes(
    base_resolution: int,
    lam_resolution: int,
    area_mask_builder: KNNAreaMaskBuilder | None = None,
) -> tuple[nx.DiGraph, np.ndarray, list[int]]:
    """Creates a global mesh with 2 levels of resolution.

    The base resolution is used to define the nodes outside the Area Of Interest (AOI),
    while the lam_resolution is used to define the nodes inside the AOI.

    Parameters
    ---------
    base_resolution : int
        Global resolution level.
    lam_resolution : int
        Local resolution level.
    area_mask_builder : KNNAreaMaskBuilder
        NearestNeighbors with the cloud of points to limit the mesh area.

    Returns
    -------
    nx_graph : nx.DiGraph
        The graph with the added nodes.
    coords_rad : np.ndarray
        The node coordinates (not ordered) in radians.
    node_ordering : list[int]
        Order of the node coordinates to be sorted by latitude and longitude.
    """
    assert area_mask_builder is not None, "AOI mask builder must be provided to build refined grid."
    # Get the low resolution nodes outside the AOI
    base_coords_rad = get_latlon_coords_icosphere(base_resolution)
    base_area_mask = ~area_mask_builder.get_mask(base_coords_rad)

    # Get the high resolution nodes inside the AOI
    lam_coords_rad = get_latlon_coords_icosphere(lam_resolution)
    lam_area_mask = area_mask_builder.get_mask(lam_coords_rad)

    coords_rad = np.concatenate([base_coords_rad[base_area_mask], lam_coords_rad[lam_area_mask]])

    node_ordering = get_coordinates_ordering(coords_rad)

    # Creates the graph, with the nodes sorted by latitude and longitude.
    nx_graph = create_nx_graph_from_tri_coords(coords_rad, node_ordering)

    return nx_graph, coords_rad, list(node_ordering)


def get_latlon_coords_icosphere(resolution: int) -> np.ndarray:
    """Get the latitude and longitude coordinates (in radians) of the icosphere.

    Parameters
    ----------
    resolution : int
        The resolution level of the icosphere.

    Returns
    -------
    np.ndarray
        The latitude and longitude coordinates, in radians, of the icosphere.
    """
    sphere = trimesh.creation.icosphere(subdivisions=resolution, radius=1.0)
    coords_rad = cartesian_to_latlon_rad(sphere.vertices)
    return coords_rad


def create_nx_graph_from_tri_coords(coords_rad: np.ndarray, node_ordering: np.ndarray) -> nx.DiGraph:
    """Creates the networkx graph from the coordinates and the node ordering."""
    graph = nx.DiGraph()
    for i, coords in enumerate(coords_rad[node_ordering]):
        node_id = node_ordering[i]
        graph.add_node(node_id, hcoords_rad=coords)

    assert list(graph.nodes.keys()) == list(node_ordering), "Nodes are not correctly added to the graph."
    assert graph.number_of_nodes() == len(node_ordering), "The number of nodes must be the same."
    return graph


def add_edges_to_nx_graph(
    graph: nx.DiGraph,
    resolutions: list[int],
    x_hops: int = 1,
    area_mask_builder: KNNAreaMaskBuilder | None = None,
) -> nx.DiGraph:
    """Adds the edges to the graph.

    This method adds multi-scale connections to the existing graph. The corresponfing nodes or vertices
    are defined by an isophere at the different esolutions (or refinement levels) specified.

    Parameters
    ----------
    graph : nx.DiGraph
        The graph to add the edges. It should correspond to the mesh nodes, without edges.
    resolutions : list[int]
        Levels of mesh refinement levels to consider.
    x_hops : int, optional
        Number of hops between 2 nodes to consider them neighbours, by default 1.
    area_mask_builder : KNNAreaMaskBuilder
        NearestNeighbors with the cloud of points to limit the mesh area, by default None.

    Returns
    -------
    graph : nx.DiGraph
        The graph with the added edges.
    """
    assert x_hops > 0, "x_hops == 0, graph would have no edges ..."

    graph_vertices = np.array([graph.nodes[i]["hcoords_rad"] for i in sorted(graph.nodes)])
    tree = BallTree(graph_vertices, metric="haversine")

    # Build the multi-scale connections
    for resolution in resolutions:
        # Define the coordinates of the isophere vertices at specified 'resolution' level
        r_sphere = trimesh.creation.icosphere(subdivisions=resolution, radius=1.0)
        r_vertices_rad = cartesian_to_latlon_rad(r_sphere.vertices)

        # Limit area of mesh points.
        if area_mask_builder is not None:
            area_mask = area_mask_builder.get_mask(r_vertices_rad)
            valid_nodes = np.where(area_mask)[0]
        else:
            valid_nodes = None

        node_neighbours = get_neighbours_within_hops(r_sphere, x_hops, valid_nodes=valid_nodes)

        _, vertex_mapping_index = tree.query(r_vertices_rad, k=1)
        neighbour_pairs = create_node_neighbours_list(graph, node_neighbours, vertex_mapping_index)
        graph.add_edges_from(neighbour_pairs)
    return graph


def get_neighbours_within_hops(
    tri_mesh: trimesh.Trimesh, x_hops: int, valid_nodes: list[int] | None = None
) -> dict[int, set[int]]:
    """Get the neigbour connections in the graph.

    Parameters
    ----------
    tri_mesh : trimesh.Trimesh
        The mesh to consider.
    x_hops : int
        Number of hops between 2 nodes to consider them neighbours.
    valid_nodes : list[int], optional
        List of valid nodes to consider, by default None. It is useful to consider only a subset of the nodes to save
        computation time.

    Returns
    -------
    neighbours : dict[int, set[int]]
        A list with the neighbours for each vertex. The element at position 'i' correspond to the neighbours to the
        i-th vertex of the mesh.
    """
    edges = tri_mesh.edges_unique

    if valid_nodes is not None:
        edges = edges[np.isin(tri_mesh.edges_unique, valid_nodes).all(axis=1)]
    else:
        valid_nodes = list(range(len(tri_mesh.vertices)))
    graph = nx.from_edgelist(edges)

    neighbours = {
        i: set(nx.ego_graph(graph, i, radius=x_hops, center=False) if i in graph else []) for i in valid_nodes
    }

    return neighbours


def add_neigbours_edges(
    graph: nx.Graph,
    node_idx: int,
    neighbour_indices: Iterable[int],
    self_loops: bool = False,
    vertex_mapping_index: np.ndarray | None = None,
) -> nx.Graph:
    """Adds the edges of one node to its neighbours.

    Parameters
    ----------
    graph : nx.Graph
        The graph.
    node_idx : int
        The node considered.
    neighbour_indices : list[int]
        The neighbours of the node.
    self_loops : bool, optional
        Whether is supported to add self-loops, by default False.
    vertex_mapping_index : np.ndarray, optional
        Index to map the vertices from the refined sphere to the original one, by default None.

    Returns
    -------
    nx.Graph
        The graph with the added edges.
    """
    graph_nodes_idx = list(sorted(graph.nodes))
    for neighbour_idx in neighbour_indices:
        if not self_loops and node_idx == neighbour_idx:  # no self-loops
            continue

        if vertex_mapping_index is not None:
            # Use the same method to add edge in all spheres
            node_neighbour = graph_nodes_idx[vertex_mapping_index[neighbour_idx][0]]
            node = graph_nodes_idx[vertex_mapping_index[node_idx][0]]
        else:
            node_neighbour = graph_nodes_idx[neighbour_idx]
            node = graph_nodes_idx[node_idx]

        # add edge to the graph
        if node in graph and node_neighbour in graph:
            graph.add_edge(node_neighbour, node)

    return graph


def create_node_neighbours_list(
    graph: nx.Graph,
    node_neighbours: dict[int, set[int]],
    vertex_mapping_index: np.ndarray | None = None,
    self_loops: bool = False,
) -> list[tuple]:
    """Preprocesses the dict of node neighbours.

    Parameters:
    -----------
    graph: nx.Graph
        The graph.
    node_neighbours: dict[int, set[int]]
        dictionairy with key: node index and value: set of neighbour node indices
    vertex_mapping_index: np.ndarry
        Index to map the vertices from the refined sphere to the original one, by default None.
    self_loops: bool
        Whether is supported to add self-loops, by default False.

    Returns:
    --------
    list: tuple
        A list with containing node neighbour pairs in tuples
    """
    graph_nodes_idx = list(sorted(graph.nodes))

    if vertex_mapping_index is None:
        vertex_mapping_index = np.arange(len(graph.nodes)).reshape(len(graph.nodes), 1)

    neighbour_pairs = [
        (graph_nodes_idx[vertex_mapping_index[node_neighbour][0]], graph_nodes_idx[vertex_mapping_index[node][0]])
        for node, neighbours in node_neighbours.items()
        for node_neighbour in neighbours
        if node != node_neighbour or (self_loops and node == node_neighbour)
    ]

    return neighbour_pairs



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging

import torch

LOGGER = logging.getLogger(__name__)


class NormaliserMixin:
    """Mixin class for normalising attributes."""

    def normalise(self, values: torch.Tensor) -> torch.Tensor:
        """Normalise the given values.

        It supports different normalisation methods: None, 'l1',
        'l2', 'unit-max', 'unit-range' and 'unit-std'.

        Parameters
        ----------
        values : np.ndarray of shape (N, M)
            Values to normalise.

        Returns
        -------
        np.ndarray
            Normalised values.
        """
        if self.norm is None:
            LOGGER.debug(f"{self.__class__.__name__} values are not normalised.")
            return values
        if self.norm == "l1":
            return values / torch.sum(values)
        if self.norm == "l2":
            return values / torch.norm(values)
        if self.norm == "unit-max":
            return values / torch.amax(values)
        if self.norm == "unit-range":
            return (values - torch.amin(values)) / (torch.amax(values) - torch.amin(values))
        if self.norm == "unit-std":
            std = torch.std(values)
            if std == 0:
                LOGGER.warning(f"Std. dev. of the {self.__class__.__name__} values is 0. Normalisation is skipped.")
                return values
            return values / std
        raise ValueError(
            f"Attribute normalisation \"{self.norm}\" is not valid. Options are: 'l1', 'l2', 'unit-max' or 'unit-std'."
        )



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from .builders.cutoff import CutOffEdges
from .builders.icon import ICONTopologicalDecoderEdges
from .builders.icon import ICONTopologicalEncoderEdges
from .builders.icon import ICONTopologicalProcessorEdges
from .builders.knn import KNNEdges
from .builders.multi_scale import MultiScaleEdges

__all__ = [
    "KNNEdges",
    "CutOffEdges",
    "MultiScaleEdges",
    "ICONTopologicalProcessorEdges",
    "ICONTopologicalEncoderEdges",
    "ICONTopologicalDecoderEdges",
]



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from __future__ import annotations

import logging
import time
from abc import ABC
from abc import abstractmethod

import numpy as np
import torch
from hydra.utils import instantiate
from torch_geometric.data import HeteroData

from anemoi.graphs.utils import get_grid_reference_distance
from anemoi.utils.config import DotDict

LOGGER = logging.getLogger(__name__)


class BaseNodeBuilder(ABC):
    """Base class for node builders.

    The node coordinates are stored in the `x` attribute of the nodes and they are stored in radians.

    Attributes
    ----------
    name : str
        name of the nodes, key for the nodes in the HeteroData graph object.
    area_mask_builder : KNNAreaMaskBuilder
        The area of interest mask builder, if any. Defaults to None.
    """

    hidden_attributes: set[str] = set()

    def __init__(self, name: str) -> None:
        self.name = name
        self.area_mask_builder = None

    def register_nodes(self, graph: HeteroData) -> HeteroData:
        """Register nodes in the graph.

        Parameters
        ----------
        graph : HeteroData
            The graph to register the nodes.

        Returns
        -------
        HeteroData
            The graph with the registered nodes.
        """
        graph[self.name].x = self.get_coordinates().to(torch.float32)
        graph[self.name].node_type = type(self).__name__

        if graph[self.name].num_nodes >= 2:
            # At least 2 nodes are needed to compute the grid_reference_distance
            graph[self.name]["_grid_reference_distance"] = get_grid_reference_distance(graph[self.name].x.cpu())
        else:
            LOGGER.warning(f"{self.__class__.__name__} registered {graph[self.name].num_nodes} nodes.")

        return graph

    def register_attributes(self, graph: HeteroData, config: DotDict | None = None) -> HeteroData:
        """Register attributes in the nodes of the graph specified.

        Parameters
        ----------
        graph : HeteroData
            The graph to register the attributes.
        config : DotDict
            The configuration of the attributes.

        Returns
        -------
        HeteroData
            The graph with the registered attributes.
        """
        for hidden_attr in self.hidden_attributes:
            graph[self.name][f"_{hidden_attr}"] = getattr(self, hidden_attr)

        for attr_name, attr_config in config.items():
            graph[self.name][attr_name] = instantiate(attr_config).compute(graph, self.name)

        return graph

    @abstractmethod
    def get_coordinates(self) -> torch.Tensor: ...

    def reshape_coords(
        self, latitudes: np.ndarray | torch.Tensor, longitudes: np.ndarray | torch.Tensor
    ) -> torch.Tensor:
        """Reshape latitude and longitude coordinates.

        Parameters
        ----------
        latitudes : np.ndarray of shape (num_nodes, )
            Latitude coordinates, in degrees.
        longitudes : np.ndarray of shape (num_nodes, )
            Longitude coordinates, in degrees.

        Returns
        -------
        torch.Tensor of shape (num_nodes, 2)
            A 2D tensor with the coordinates, in radians.
        """
        if isinstance(latitudes, np.ndarray):
            latitudes = torch.from_numpy(latitudes)

        if isinstance(longitudes, np.ndarray):
            longitudes = torch.from_numpy(longitudes)

        coords = torch.stack([latitudes, longitudes], axis=-1).reshape((-1, 2))
        return torch.deg2rad(coords)

    def update_graph(self, graph: HeteroData, attrs_config: DotDict | None = None) -> HeteroData:
        """Update the graph with new nodes.

        Parameters
        ----------
        graph : HeteroData
            Input graph.
        attrs_config : DotDict
            The configuration of the attributes.

        Returns
        -------
        HeteroData
            The graph with new nodes included.
        """
        t0 = time.time()
        graph = self.register_nodes(graph)
        t1 = time.time()
        LOGGER.debug("Time to register node coordinates (%s): %.2f s", self.__class__.__name__, t1 - t0)

        t0 = time.time()
        graph = self.register_attributes(graph, attrs_config or {})
        t1 = time.time()
        LOGGER.debug("Time to register node coordinates (%s): %.2f s", self.__class__.__name__, t1 - t0)

        return graph



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging
from pathlib import Path
from typing import Literal
from typing import Optional
from typing import Union

import matplotlib.pyplot as plt
import numpy as np
import torch
from torch_geometric.data import HeteroData
from torch_geometric.data.storage import EdgeStorage
from torch_geometric.data.storage import NodeStorage

from anemoi.graphs.plotting.prepare import compute_node_adjacencies
from anemoi.graphs.plotting.prepare import get_edge_attribute_dims
from anemoi.graphs.plotting.prepare import get_node_attribute_dims

LOGGER = logging.getLogger(__name__)


def plot_distribution_node_attributes(graph: HeteroData, out_file: Optional[Union[str, Path]] = None) -> None:
    """Figure with the distribution of the node attributes.

    Each row represents a node type and each column an attribute dimension.
    """
    num_nodes = len(graph.node_types)
    attr_dims = get_node_attribute_dims(graph)
    plot_distribution_attributes(graph.node_items(), num_nodes, attr_dims, "Node", out_file)


def plot_distribution_edge_attributes(graph: HeteroData, out_file: Optional[Union[str, Path]] = None) -> None:
    """Figure with the distribution of the edge attributes.

    Each row represents a edge type and each column an attribute dimension.
    """
    num_edges = len(graph.edge_types)
    attr_dims = get_edge_attribute_dims(graph)
    plot_distribution_attributes(graph.edge_items(), num_edges, attr_dims, "Edge", out_file)


def plot_distribution_node_derived_attributes(graph, outfile: Optional[Union[str, Path]] = None):
    """Figure with the distribution of the node derived attributes.

    Each row represents a node type and each column an attribute dimension.
    """
    node_adjs = {}
    node_attr_dims = {}
    for source_name, _, target_name in graph.edge_types:
        node_adj_tensor = compute_node_adjacencies(graph, source_name, target_name)
        node_adj_tensor = torch.from_numpy(node_adj_tensor.reshape((node_adj_tensor.shape[0], -1)))
        node_adj_key = f"# edges from {source_name}"

        # Store node adjacencies
        if target_name in node_adjs:
            node_adjs[target_name] = node_adjs[target_name] | {node_adj_key: node_adj_tensor}
        else:
            node_adjs[target_name] = {node_adj_key: node_adj_tensor}

        # Store attribute dimension
        if node_adj_key not in node_attr_dims:
            node_attr_dims[node_adj_key] = node_adj_tensor.shape[1]

    node_adj_list = [(k, v) for k, v in node_adjs.items()]

    plot_distribution_attributes(node_adj_list, len(node_adjs), node_attr_dims, "Node", outfile)


def plot_distribution_attributes(
    graph_items: Union[NodeStorage, EdgeStorage],
    num_items: int,
    attr_dims: dict,
    item_type: Literal["Edge", "Node"],
    out_file: Optional[Union[str, Path]] = None,
) -> None:
    """Figure with the distribution of the node and edge attributes.

    Each row represents a node or edge type and each column an attribute dimension.
    """
    dim_attrs = sum(attr_dims.values())

    if dim_attrs == 0:
        LOGGER.warning("No edge attributes found in the graph.")
        return None

    # Define the layout
    _, axs = plt.subplots(num_items, dim_attrs, figsize=(10 * num_items, 10))
    if num_items == dim_attrs == 1:
        axs = np.array([[axs]])
    elif axs.ndim == 1:
        axs = axs.reshape(num_items, dim_attrs)

    for i, (item_name, item_store) in enumerate(graph_items):
        for j, (attr_name, attr_values) in enumerate(attr_dims.items()):
            for dim in range(attr_values):
                if attr_name in item_store:
                    axs[i, j + dim].hist(item_store[attr_name][:, dim].float(), bins=50)

                    axs[i, j + dim].set_ylabel("".join(item_name).replace("to", " --> "))
                    axs[i, j + dim].set_title(attr_name if attr_values == 1 else f"{attr_name}_{dim}")
                    if i == num_items - 1:
                        axs[i, j + dim].set_xlabel(attr_name if attr_values == 1 else f"{attr_name}_{dim}")
                else:
                    axs[i, j + dim].set_axis_off()

    plt.suptitle(f"{item_type} Attributes distribution", fontsize=14)
    plt.savefig(out_file)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from __future__ import annotations

import logging

import numpy as np
import torch
from scipy.sparse import coo_matrix
from torch_geometric.data.storage import NodeStorage

from anemoi.graphs.generate.transforms import latlon_rad_to_cartesian

LOGGER = logging.getLogger(__name__)


class NodeMaskingMixin:
    """Mixin class for masking source/target nodes when building edges."""

    def get_node_coordinates(
        self, source_nodes: NodeStorage, target_nodes: NodeStorage
    ) -> tuple[torch.Tensor, torch.Tensor]:
        """Get the node coordinates."""
        source_coords, target_coords = source_nodes.x, target_nodes.x

        if self.source_mask_attr_name is not None:
            source_coords = source_coords[source_nodes[self.source_mask_attr_name].squeeze()]

        if self.target_mask_attr_name is not None:
            target_coords = target_coords[target_nodes[self.target_mask_attr_name].squeeze()]

        return source_coords.to(self.device), target_coords.to(self.device)

    def get_cartesian_node_coordinates(
        self, source_nodes: NodeStorage, target_nodes: NodeStorage
    ) -> tuple[torch.Tensor, torch.Tensor]:
        source_coords, target_coords = self.get_node_coordinates(source_nodes, target_nodes)
        source_coords = latlon_rad_to_cartesian(source_coords)
        target_coords = latlon_rad_to_cartesian(target_coords)
        return source_coords, target_coords

    def undo_masking(self, adj_matrix, source_nodes: NodeStorage, target_nodes: NodeStorage):
        if self.target_mask_attr_name is not None:
            target_mask = target_nodes[self.target_mask_attr_name].squeeze().cpu()
            assert adj_matrix.shape[0] == target_mask.sum()
            target_mapper = dict(zip(list(range(adj_matrix.shape[0])), np.where(target_mask)[0]))
            adj_matrix.row = np.vectorize(target_mapper.get)(adj_matrix.row)

        if self.source_mask_attr_name is not None:
            source_mask = source_nodes[self.source_mask_attr_name].squeeze().cpu()
            assert adj_matrix.shape[1] == source_mask.sum()
            source_mapper = dict(zip(list(range(adj_matrix.shape[1])), np.where(source_mask)[0]))
            adj_matrix.col = np.vectorize(source_mapper.get)(adj_matrix.col)

        if self.source_mask_attr_name is not None or self.target_mask_attr_name is not None:
            true_shape = target_nodes.x.shape[0], source_nodes.x.shape[0]
            adj_matrix = coo_matrix((adj_matrix.data, (adj_matrix.row, adj_matrix.col)), shape=true_shape)

        return adj_matrix



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import pytest
from torch_geometric.data import HeteroData

from anemoi.graphs.edges import CutOffEdges


def test_init():
    """Test CutOffEdges initialization."""
    CutOffEdges("test_nodes1", "test_nodes2", 0.5)


@pytest.mark.parametrize("cutoff_factor", [-0.5, "hello", None])
def test_fail_init(cutoff_factor: str):
    """Test CutOffEdges initialization with invalid cutoff."""
    with pytest.raises(AssertionError):
        CutOffEdges("test_nodes1", "test_nodes2", cutoff_factor)


def test_cutoff(graph_with_nodes: HeteroData):
    """Test CutOffEdges."""
    builder = CutOffEdges("test_nodes", "test_nodes", 0.5)
    graph = builder.update_graph(graph_with_nodes)
    assert ("test_nodes", "to", "test_nodes") in graph.edge_types



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from __future__ import annotations

import logging
from abc import ABC
from abc import abstractmethod
from typing import Any

import torch
from hydra.utils import instantiate
from torch_geometric.data import HeteroData

from anemoi.graphs import EARTH_RADIUS
from anemoi.graphs.edges.attributes import EdgeLength
from anemoi.graphs.utils import NodesAxis
from anemoi.graphs.utils import get_edge_attributes

LOGGER = logging.getLogger(__name__)


class PostProcessor(ABC):

    @abstractmethod
    def update_graph(self, graph: HeteroData, **kwargs: Any) -> HeteroData:
        raise NotImplementedError(f"The {self.__class__.__name__} class does not implement the method update_graph().")


class BaseNodeMaskingProcessor(PostProcessor, ABC):
    """Base class for mask based node processor."""

    def __init__(
        self,
        nodes_name: str,
        save_mask_indices_to_attr: str | None = None,
    ) -> None:
        self.nodes_name = nodes_name
        self.save_mask_indices_to_attr = save_mask_indices_to_attr
        self.mask: torch.Tensor = None

    def removing_nodes(self, graph: HeteroData) -> HeteroData:
        """Remove nodes based on the mask passed."""
        for attr_name in graph[self.nodes_name].node_attrs():
            graph[self.nodes_name][attr_name] = graph[self.nodes_name][attr_name][self.mask]

        return graph

    def create_indices_mapper_from_mask(self) -> dict[int, int]:
        return dict(zip(torch.where(self.mask)[0].tolist(), list(range(self.mask.sum()))))

    def update_edge_indices(self, graph: HeteroData) -> HeteroData:
        """Update the edge indices to the new position of the nodes."""
        idx_mapping = self.create_indices_mapper_from_mask()
        for edges_name in graph.edge_types:
            if edges_name[0] == self.nodes_name:
                graph[edges_name].edge_index[0] = graph[edges_name].edge_index[0].cpu().apply_(idx_mapping.get)

            if edges_name[2] == self.nodes_name:
                graph[edges_name].edge_index[1] = graph[edges_name].edge_index[1].cpu().apply_(idx_mapping.get)

        return graph

    @abstractmethod
    def compute_mask(self, graph: HeteroData) -> torch.Tensor: ...

    def add_attribute(self, graph: HeteroData) -> HeteroData:
        """Add an attribute of the mask indices as node attribute."""
        if self.save_mask_indices_to_attr is not None:
            LOGGER.info(
                f"An attribute {self.save_mask_indices_to_attr} has been added with the indices to mask the nodes from the original graph."
            )
            mask_indices = torch.where(self.mask)[0].reshape((graph[self.nodes_name].num_nodes, -1))
            graph[self.nodes_name][self.save_mask_indices_to_attr] = mask_indices

        return graph

    def update_graph(self, graph: HeteroData, **kwargs: Any) -> HeteroData:
        """Post-process the graph.

        Parameters
        ----------
        graph: HeteroData
            The graph to post-process.
        kwargs: Any
            Additional keyword arguments.

        Returns
        -------
        HeteroData
            The post-processed graph.
        """
        self.mask = self.compute_mask(graph)
        LOGGER.info(f"Removing {(~self.mask).sum()} nodes from {self.nodes_name}.")
        graph = self.removing_nodes(graph)
        graph = self.update_edge_indices(graph)
        graph = self.add_attribute(graph)
        return graph


class RemoveUnconnectedNodes(BaseNodeMaskingProcessor):
    """Remove unconnected nodes in the graph.

    Attributes
    ----------
    nodes_name: str
        Name of the unconnected nodes to remove.
    ignore: str, optional
        Name of an attribute to ignore when removing nodes. Nodes with
        this attribute set to True will not be removed.
    save_mask_indices_to_attr: str, optional
        Name of the attribute to save the mask indices. If provided,
        the indices of the kept nodes will be saved in this attribute.

    Methods
    -------
    compute_mask(graph)
        Compute the mask of the connected nodes.
    prune_graph(graph, mask)
        Prune the nodes with the specified mask.
    add_attribute(graph, mask)
        Add an attribute of the mask indices as node attribute.
    update_graph(graph)
        Post-process the graph.
    """

    def __init__(
        self,
        nodes_name: str,
        save_mask_indices_to_attr: str | None = None,
        ignore: str | None = None,
    ) -> None:
        super().__init__(nodes_name, save_mask_indices_to_attr)
        self.ignore = ignore

    def compute_mask(self, graph: HeteroData) -> torch.Tensor:
        """Compute the mask of connected nodes."""
        nodes = graph[self.nodes_name]
        connected_mask = torch.zeros(nodes.num_nodes, dtype=torch.bool)

        if self.ignore is not None:
            LOGGER.info(f"The nodes with {self.ignore}=True will not be removed.")
            connected_mask[nodes[self.ignore].bool().squeeze()] = True

        for (source_name, _, target_name), edges in graph.edge_items():
            if source_name == self.nodes_name:
                connected_mask[edges.edge_index[0]] = True

            if target_name == self.nodes_name:
                connected_mask[edges.edge_index[1]] = True

        return connected_mask


class BaseSortEdgeIndex(PostProcessor, ABC):
    """Base class for sort edge indices processor."""

    nodes_axis: NodesAxis | None = None

    def __init__(self, descending: bool = True) -> None:
        assert self.nodes_axis is not None, f"{self.__class__.__name__} must define the nodes_axis class attribute."
        self.descending = descending

    def get_sorting_mask(self, edges: dict) -> torch.Tensor:
        sort_indices = torch.sort(edges["edge_index"], descending=self.descending, dim=1)
        return sort_indices.indices[self.nodes_axis.value]

    @staticmethod
    def get_edge_dim(edge_attr: str) -> int:
        # edge_index has shape (2, n_edges) and other edge attributes have (n_edges, attr_dim)
        return int(edge_attr == "edge_index")

    @staticmethod
    def sort_by_indices(x: torch.Tensor, indices: torch.Tensor, dim: int = 1) -> torch.Tensor:
        return x.index_select(dim=dim, index=indices)

    def update_graph(self, graph: HeteroData) -> HeteroData:
        """Sort all edge indices in the graph.

        Parameters
        ----------
        graph: HeteroData
            The graph to post-process.

        Returns
        -------
        HeteroData
            The post-processed graph.
        """
        for (src, to, dst), edges in graph.edge_items():
            sort_indices = self.get_sorting_mask(edges)
            for edge_attr_name in edges.edge_attrs():
                dim = BaseSortEdgeIndex.get_edge_dim(edge_attr_name)
                edge_attr = BaseSortEdgeIndex.sort_by_indices(edges[edge_attr_name], sort_indices, dim=dim)
                graph[(src, to, dst)][edge_attr_name] = edge_attr
        return graph


class SortEdgeIndexBySourceNodes(BaseSortEdgeIndex):
    nodes_axis = NodesAxis.SOURCE


class SortEdgeIndexByTargetNodes(BaseSortEdgeIndex):
    nodes_axis = NodesAxis.TARGET


class BaseEdgeMaskingProcessor(PostProcessor, ABC):
    """Base class for mask based edge processor.

    Attributes
    ----------
    source_name: str
        Name of the source nodes of edges to remove.
    target_name: str
        Name of the target nodes of edges to remove.
    """

    def __init__(
        self,
        source_name: str,
        target_name: str,
    ) -> None:
        self.source_name = source_name
        self.target_name = target_name
        self.edges_name = (self.source_name, "to", self.target_name)
        self.mask: torch.Tensor = None

    def removing_edges(self, graph: HeteroData) -> HeteroData:
        """Remove edges based on the mask passed."""
        for attr_name in graph[self.edges_name].edge_attrs():
            if attr_name == "edge_index":
                graph[self.edges_name][attr_name] = graph[self.edges_name][attr_name][:, self.mask]
            else:
                graph[self.edges_name][attr_name] = graph[self.edges_name][attr_name][self.mask, :]

        return graph

    @abstractmethod
    def compute_mask(self, graph: HeteroData) -> torch.Tensor: ...

    def recompute_attributes(self, graph: HeteroData, graph_config: dict) -> HeteroData:
        """Recompute attributes"""
        edge_attributes = get_edge_attributes(graph_config, self.source_name, self.target_name)
        for attr_name, edge_attr_builder in edge_attributes.items():
            LOGGER.info(f"Recomputing edge attribute {attr_name}.")
            graph[self.edges_name][attr_name] = instantiate(edge_attr_builder)(
                x=(graph[self.source_name], graph[self.target_name]), edge_index=graph[self.edges_name].edge_index
            )
        return graph

    def update_graph(self, graph: HeteroData, **kwargs: Any) -> HeteroData:
        """Post-process the graph.

        Parameters
        ----------
        graph: HeteroData
            The graph to post-process.
        kwargs: Any
            Additional keyword arguments.

        Returns
        -------
        HeteroData
            The post-processed graph.
        """
        self.mask = self.compute_mask(graph)
        LOGGER.info(f"Removing {(~self.mask).sum()} edges from {self.edges_name}.")
        graph = self.removing_edges(graph)
        graph_config = kwargs.get("graph_config", {})
        graph = self.recompute_attributes(graph, graph_config)
        return graph


class RestrictEdgeLength(BaseEdgeMaskingProcessor):
    """Remove edges longer than a given treshold from the graph.

    Attributes
    ----------
    source_name: str
        Name of the source nodes of edges to remove.
    target_name: str
        Name of the target nodes of edges to remove.
    max_length_km: float
        The maximal length of edges not to be removed.
    source_mask_attr_name: str , optional
         the postprocessing will be restricted to edges with source node having True in this mask_attr
    target_mask_attr_name: str, optional
        the postprocessing will be restricted to edges with target node having True in this mask_attr
    Methods
    -------
    compute_mask(graph)
        Compute the mask of the relevant edges longer than max_length_km.
    """

    def __init__(
        self,
        source_name: str,
        target_name: str,
        max_length_km: float,
        source_mask_attr_name: str | None = None,
        target_mask_attr_name: str | None = None,
    ) -> None:
        super().__init__(source_name, target_name)
        self.treshold = max_length_km
        self.source_mask_attr_name = source_mask_attr_name
        self.target_mask_attr_name = target_mask_attr_name

    def compute_mask(self, graph: HeteroData) -> torch.Tensor:
        source_nodes = graph[self.source_name]
        target_nodes = graph[self.target_name]
        edge_index = graph[self.edges_name].edge_index
        lengths = EARTH_RADIUS * EdgeLength()(x=(source_nodes, target_nodes), edge_index=edge_index)
        mask = torch.where(lengths > self.treshold, False, True).squeeze()
        cases = [
            (self.source_mask_attr_name, source_nodes, 0),
            (self.target_mask_attr_name, target_nodes, 1),
        ]
        for mask_attr_name, nodes, i in cases:
            if mask_attr_name:
                attr_mask = nodes[mask_attr_name].squeeze()
                edge_mask = attr_mask[edge_index[i]]
                mask = torch.logical_or(mask, ~edge_mask)
        return mask



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from __future__ import annotations

import numpy as np
import torch
from torch_geometric.data import HeteroData

from anemoi.graphs.generate.icon_mesh import ICONCellDataGrid
from anemoi.graphs.generate.icon_mesh import ICONMultiMesh
from anemoi.graphs.nodes.builders.base import BaseNodeBuilder
from anemoi.utils.config import DotDict


class ICONNodes(BaseNodeBuilder):
    """ICON grid (cell and vertex locations)."""

    def __init__(self, name: str, grid_filename: str, max_level_multimesh: int, max_level_dataset: int) -> None:
        self.grid_filename = grid_filename

        self.multi_mesh = ICONMultiMesh(self.grid_filename, max_level=max_level_multimesh)
        self.cell_grid = ICONCellDataGrid(self.grid_filename, self.multi_mesh, max_level=max_level_dataset)

        super().__init__(name)

    def get_coordinates(self) -> torch.Tensor:
        return torch.from_numpy(self.multi_mesh.nodeset.gc_vertices.astype(np.float32)).fliplr()

    def register_attributes(self, graph: HeteroData, config: DotDict) -> HeteroData:
        graph[self.name]["_grid_filename"] = self.grid_filename
        graph[self.name]["_multi_mesh"] = self.multi_mesh
        graph[self.name]["_cell_grid"] = self.cell_grid
        return super().register_attributes(graph, config)


class ICONTopologicalBaseNodeBuilder(BaseNodeBuilder):
    """Base class for data mesh or processor mesh based on an ICON grid.

    Parameters
    ----------
    name : str
        key for the nodes in the HeteroData graph object.
    icon_mesh : str
        key corresponding to the ICON mesh (cells and vertices).
    """

    def __init__(self, name: str, icon_mesh: str) -> None:
        self.icon_mesh = icon_mesh
        super().__init__(name)

    def update_graph(self, graph: HeteroData, attrs_config: DotDict | None = None) -> HeteroData:
        """Update the graph with new nodes."""
        self.icon_sub_graph = graph[self.icon_mesh][self.sub_graph_address]
        return super().update_graph(graph, attrs_config)


class ICONMultimeshNodes(ICONTopologicalBaseNodeBuilder):
    """Processor mesh based on an ICON grid."""

    def __init__(self, name: str, icon_mesh: str) -> None:
        self.sub_graph_address = "_multi_mesh"
        super().__init__(name, icon_mesh)

    def get_coordinates(self) -> torch.Tensor:
        return torch.from_numpy(self.icon_sub_graph.nodeset.gc_vertices.astype(np.float32)).fliplr()


class ICONCellGridNodes(ICONTopologicalBaseNodeBuilder):
    """Data mesh based on an ICON grid."""

    def __init__(self, name: str, icon_mesh: str) -> None:
        self.sub_graph_address = "_cell_grid"
        super().__init__(name, icon_mesh)

    def get_coordinates(self) -> torch.Tensor:
        return torch.from_numpy(self.icon_sub_graph.nodeset[0].gc_vertices.astype(np.float32)).fliplr()



# (C) Copyright 2024-2025 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


from __future__ import annotations

from typing import Annotated
from typing import Literal
from typing import Union

from pydantic import Field
from pydantic import PositiveFloat
from pydantic import PositiveInt

from anemoi.utils.schemas import BaseModel


class KNNEdgeSchema(BaseModel):
    target_: Literal["anemoi.graphs.edges.KNNEdges"] = Field("anemoi.graphs.edges.KNNEdges", alias="_target_")
    "KNN based edges implementation from anemoi.graphs.edges."
    num_nearest_neighbours: PositiveInt = Field(example=3)
    "Number of nearest neighbours. Default to 3."
    source_mask_attr_name: Union[str, None] = Field(default=None, examples=["boundary_mask"])
    "Mask to apply to source nodes of the edges. Default to None."
    target_mask_attr_name: Union[str, None] = Field(default=None, examples=["boundary_mask"])
    "Mask to apply to target nodes of the edges. Default to None."


class CutoffEdgeSchema(BaseModel):
    target_: Literal["anemoi.graphs.edges.CutOffEdges"] = Field("anemoi.graphs.edges.CutOffEdges", alias="_target_")
    "Cut-off based edges implementation from anemoi.graphs.edges."
    cutoff_factor: PositiveFloat = Field(example=0.6)
    "Factor to multiply the grid reference distance to get the cut-off radius. Default to 0.6."
    source_mask_attr_name: Union[str, None] = Field(default=None, examples=["boundary_mask"])
    "Mask to apply to source nodes of the edges. Default to None."
    target_mask_attr_name: Union[str, None] = Field(default=None, examples=["boundary_mask"])
    "Mask to apply to target nodes of the edges. Default to None."


class MultiScaleEdgeSchema(BaseModel):
    target_: Literal["anemoi.graphs.edges.MultiScaleEdges"] = Field(
        "anemoi.graphs.edges.MultiScaleEdges",
        alias="_target_",
    )
    "Multi-casle edges implementation from anemoi.graphs.edges."
    x_hops: PositiveInt = Field(example=1)
    "Number of hops (in the refined icosahedron) between two nodes to connect them with an edge. Default to 1."
    scale_resolutions: Union[PositiveInt, list[PositiveInt], None] = Field(examples=[1, 2, 3, 4, 5])
    "Specifies the resolution scales for computing the hop neighbourhood."
    source_mask_attr_name: Union[str, None] = Field(default=None, examples=["boundary_mask"])
    "Mask to apply to source nodes of the edges. Default to None."
    target_mask_attr_name: Union[str, None] = Field(default=None, examples=["boundary_mask"])
    "Mask to apply to target nodes of the edges. Default to None."


class ICONTopologicalEdgeSchema(BaseModel):
    target_: Literal[
        "anemoi.graphs.edges.ICONTopologicalProcessorEdges",
        "anemoi.graphs.edges.ICONTopologicalEncoderEdges",
        "anemoi.graphs.edges.ICONTopologicalDecoderEdges",
    ] = Field("anemoi.graphs.edges.ICONTopologicalProcessorEdges", alias="_target_")
    icon_mesh: str
    "The name of the ICON mesh (defines both the processor mesh and the data)."
    source_mask_attr_name: Union[str, None] = Field(default=None, examples=["boundary_mask"])
    "Mask to apply to source nodes of the edges. Default to None."
    target_mask_attr_name: Union[str, None] = Field(default=None, examples=["boundary_mask"])
    "Mask to apply to target nodes of the edges. Default to None."


class EdgeAttributeSchema(BaseModel):
    target_: Literal["anemoi.graphs.edges.attributes.EdgeLength", "anemoi.graphs.edges.attributes.EdgeDirection"] = (
        Field("anemoi.graphs.edges.attributes.EdgeLength", alias="_target_")
    )
    "Edge attributes object from anemoi.graphs.edges."
    norm: Literal["unit-max", "l1", "l2", "unit-sum", "unit-std"] = Field(example="unit-std")
    "Normalisation method applied to the edge attribute."


EdgeBuilderSchemas = Annotated[
    Union[KNNEdgeSchema, CutoffEdgeSchema, MultiScaleEdgeSchema, ICONTopologicalEdgeSchema],
    Field(discriminator="target_"),
]



# (C) Copyright 2024- Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from __future__ import annotations

import logging

import numpy as np
import torch
from torch_geometric.data.storage import NodeStorage

from anemoi.datasets import open_dataset
from anemoi.graphs.nodes.attributes.base_attributes import BooleanBaseNodeAttribute

LOGGER = logging.getLogger(__name__)


class NonmissingAnemoiDatasetVariable(BooleanBaseNodeAttribute):
    """Mask of valid (not missing) values of a Anemoi dataset variable.

    It reads a variable from a Anemoi dataset and returns a boolean mask of nonmissing values in the first timestep.

    Attributes
    ----------
    variable : str
        Variable to read from the Anemoi dataset.

    Methods
    -------
    compute(self, graph, nodes_name)
        Compute the attribute for each node.
    """

    def __init__(self, variable: str) -> None:
        super().__init__()
        self.variable = variable

    def get_raw_values(self, nodes: NodeStorage, **kwargs) -> torch.Tensor:
        assert nodes["node_type"] in [
            "ZarrDatasetNodes",
            "AnemoiDatasetNodes",
        ], f"{self.__class__.__name__} can only be used with AnemoiDatasetNodes."
        ds = open_dataset(nodes["_dataset"], select=self.variable)[0].squeeze()
        return torch.from_numpy(~np.isnan(ds))


class CutOutMask(BooleanBaseNodeAttribute):
    """Cut out mask."""

    def get_raw_values(self, nodes: NodeStorage, **kwargs) -> torch.Tensor:
        assert "_dataset" in nodes and isinstance(
            nodes["_dataset"], dict
        ), "The '_dataset' attribute must be a dictionary."
        assert "cutout" in nodes["_dataset"], "The 'dataset' attribute must contain a 'cutout' key."
        num_lam, num_other = open_dataset(nodes["_dataset"]).grids
        return torch.tensor([True] * num_lam + [False] * num_other, dtype=torch.bool)



# (C) Copyright 2024-2025 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


from __future__ import annotations

import logging
from typing import Literal
from typing import Union

from pydantic import Field

from anemoi.utils.schemas import BaseModel

LOGGER = logging.getLogger(__name__)


class PlanarAreaWeightSchema(BaseModel):
    target_: Literal[
        "anemoi.graphs.nodes.attributes.AreaWeights",
        "anemoi.graphs.nodes.attributes.PlanarAreaWeights",
        "anemoi.graphs.nodes.attributes.UniformWeights",
    ] = Field(..., alias="_target_")
    "Implementation of the area of the nodes as the weights from anemoi.graphs.nodes.attributes."
    norm: Literal["unit-max", "l1", "l2", "unit-sum", "unit-std"] = Field(example="unit-max")
    "Normalisation of the weights."


class SphericalAreaWeightSchema(BaseModel):
    target_: Literal["anemoi.graphs.nodes.attributes.SphericalAreaWeights"] = Field(..., alias="_target_")
    "Implementation of the 3D area of the nodes as the weights from anemoi.graphs.nades.attributes."
    norm: Literal["unit-max", "l1", "l2", "unit-sum", "unit-std"] = Field(example="unit-max")
    "Normalisation of the weights."
    fill_value: float = Field(example=0)
    "Value to fill the empty regions."


class CutOutMaskSchema(BaseModel):
    target_: Literal["anemoi.graphs.nodes.attributes.CutOutMask"] = Field(..., alias="_target_")
    "Implementation of the cutout mask from anemoi.graphs.nodes.attributes."


class NonmissingAnemoiDatasetVariableSchema(BaseModel):
    target_: Literal["anemoi.graphs.nodes.attributes.NonmissingAnemoiDatasetVariable"] = Field(..., alias="_target_")
    (
        "Implementation of a mask from the nonmissing values of a anemoi-datasets variable "
        "from anemoi.graphs.nodes.attributes."
    )
    variable: str
    "The anemoi-datasets variable to use."


SingleAttributeSchema = Union[
    PlanarAreaWeightSchema,
    SphericalAreaWeightSchema,
    CutOutMaskSchema,
    NonmissingAnemoiDatasetVariableSchema,
]


class BooleanOperationSchema(BaseModel):
    target_: Literal[
        "anemoi.graphs.nodes.attributes.BooleanNot",
        "anemoi.graphs.nodes.attributes.BooleanAndMask",
        "anemoi.graphs.nodes.attributes.BooleanOrMask",
    ] = Field(..., alias="_target_")
    "Implementation of boolean masks from anemoi.graphs.nodes.attributes"
    masks: Union[str, list[str], SingleAttributeSchema, list[SingleAttributeSchema]]


NodeAttributeSchemas = Union[
    SingleAttributeSchema,
    BooleanOperationSchema,
]



# (C) Copyright 2024-2025 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


from __future__ import annotations

from enum import Enum
from typing import Literal
from typing import Union

from pydantic import Field

from anemoi.utils.schemas import BaseModel


class ImplementedEdgeAttributeSchema(str, Enum):
    edge_length = "anemoi.graphs.edges.attributes.EdgeLength"
    edge_dirs = "anemoi.graphs.edges.attributes.EdgeDirection"
    azimuth = "anemoi.graphs.edges.attributes.Azimuth"


class BaseEdgeAttributeSchema(BaseModel):
    target_: ImplementedEdgeAttributeSchema = Field(..., alias="_target_")
    "Edge attribute builder object from anemoi.graphs.edges.attributes"
    norm: Literal["unit-max", "l1", "l2", "unit-sum", "unit-std"] = Field(example="unit-std")
    "Normalisation method applied to the edge attribute."


class EdgeAttributeFromNodeSchema(BaseModel):
    target_: Literal[
        "anemoi.graphs.edges.attributes.AttributeFromSourceNode",
        "anemoi.graphs.edges.attributes.AttributeFromTargetNode",
    ] = Field(..., alias="_target_")
    "Edge attributes from node attribute"
    norm: Literal["unit-max", "l1", "l2", "unit-sum", "unit-std"] = Field(example="unit-std")
    "Normalisation method applied to the edge attribute."


EdgeAttributeSchema = Union[BaseEdgeAttributeSchema, EdgeAttributeFromNodeSchema]
